/*
 * Copyright (c) 2002-2004 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-dnstsk-0.c,v 1.42 2013/11/17 03:12:57 ca Exp $")

#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/magic.h"
#include "sm/sysexits.h"
#include "sm/memops.h"
#include "sm/test.h"
#include "sm/evthr.h"
#include "sm/dns.h"
#include "sm/dns-int.h"
#include "sm/dnstsk.h"
#include "sm/io.h"
#include "sm/misc.h"

/*
**  Should be protected by a mutex...
**  Moreover, this triggers too early if multiple callbacks are used.
*/
static bool terminate = false;

static int Verbose = 0;
static uint MaxCount = 50;
static int Timeout = 0;

/* Test program options */
#define T_OPT_USE_TCP	0x0001
#define T_OPT_TWICE	0x0002
#define T_OPT_MX	0x0004
#define T_OPT_A		0x0008
#define T_OPT_AAAA	0x0010
#define T_OPT_DEBUG	0x1000

static sm_ret_T
fcts(sm_evthr_task_P tsk)
{
	static uint count = 0;
	struct timeval sleept, delay;

	++count;
	if (Verbose > 2 && (count % 10 == 0))
		sm_io_fprintf(smioout, "fcts: count=%u, ", count);
	if (terminate || count > MaxCount)
		return EVTHR_DEL|EVTHR_TERM;
	sm_memzero(&sleept, sizeof(sleept));
	gettimeofday(&sleept, NULL);
	delay.tv_sec = 0;
	delay.tv_usec = 1000;
	timeradd(&sleept, &delay, &tsk->evthr_t_sleep);
	return EVTHR_SLPQ;
}

static sm_ret_T
callback(dns_res_P dns_res, void *ctx)
{
	dns_type_T type;
	dns_rese_P dns_rese;

	SM_TEST(dns_res != NULL);
	if (Verbose > 1)
		sm_io_fprintf(smioout, "callback: got %p, ", dns_res);
	if (NULL == dns_res) {
		terminate = true;
		return SM_FAILURE;
	}
	if (sm_is_err(dns_res->dnsres_ret)) {
		sm_io_fprintf(smioout, "error: ret=%m\n", dns_res->dnsres_ret);
		terminate = true;
		return SM_FAILURE;
	}

	if (Verbose == 0) {
		terminate = true;
		return SM_SUCCESS;
	}

	sm_io_fprintf(smioout, "dnsres: entries=%u\n", dns_res->dnsres_entries);
	sm_io_fprintf(smioout, "dnsres: max=%u\n", dns_res->dnsres_maxentries);

	for (dns_rese = DRESL_FIRST(dns_res);
	     dns_rese != DRESL_END(dns_res);
	     dns_rese = DRESL_NEXT(dns_rese))
	{
		type = dns_rese->dnsrese_type;
		sm_io_fprintf(smioout, "ttl=%u\n", dns_rese->dnsrese_ttl);
		switch (type) {
		  case T_MX:
			sm_io_fprintf(smioout, "pref=%u\n", dns_rese->dnsrese_pref);
			sm_io_fprintf(smioout, "weight=%u\n",
				dns_rese->dnsrese_weight);
			sm_io_fprintf(smioout, "value=%s\n\n",
				sm_cstr_data(dns_rese->dnsrese_val.dnsresu_name));
			break;
		  case T_A:
			sm_io_fprintf(smioout, "IPv4=%x\n",
				dns_rese->dnsrese_val.dnsresu_a);
			break;
#if MTA_NETINET6
		  case T_AAAA: {
			sm_sockaddr_T sockaddr;
/*
			struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)&dns_rese->dnsrese_val.dnsresu_aaaa;
			char ipv6str[64];
*/

			sockaddr.sa.sa_family=AF_INET6;
			IPV6_ASSIGN(&sockaddr.sin6.sin6_addr, &dns_rese->dnsrese_val.dnsresu_aaaa);
			sm_io_fprintf(smioout, "IPv6=%a\n", &sockaddr);
/*
			sm_io_fprintf(smioout, "IPv6=%s\n",
				inet_ntop(AF_INET6, &sin6->sin6_addr, ipv6str, sizeof(ipv6str)));
*/
			}
			break;
#endif
		  case T_CNAME:
			sm_io_fprintf(smioout, "CNAME=%s\n\n",
				sm_cstr_data(dns_rese->dnsrese_val.dnsresu_name));
			break;
		  default:
			sm_io_fprintf(smioout, "unknown type=%d\n", type);
			break;
		}
	}
	terminate = true;
	return SM_SUCCESS;
}

/*
**  XXX Change this to take a list of hosts? (argv[])
*/

static void
testdns(char *host, uint flags, uint32_t ipv4)
{
	sm_ret_T ret;
	int n;
	dns_mgr_ctx_P dns_mgr_ctx;
	sm_evthr_ctx_P evthr_ctx;
	sm_evthr_task_P	task, task3;
	dns_tsk_P dns_tsk;
	sm_cstr_P query;
	dns_type_T type;
	struct timeval sleept;

	if (host == NULL || *host == '\0')
		return;

	dns_tsk = NULL;
	evthr_ctx = NULL;
	task = task3 = NULL;
	dns_mgr_ctx = NULL;

	query = sm_cstr_scpyn0((const uchar *) host, strlen(host));
	SM_TEST(query != NULL);
	if (NULL == query)
		return;
	sm_memzero(&sleept, sizeof(sleept));

	ret = thr_init();
	SM_TEST(sm_is_success(ret));
	if (sm_is_err(ret))
		goto errq;
	ret = evthr_init(&evthr_ctx, 1, 6, 10);
	SM_TEST(sm_is_success(ret));
	if (sm_is_err(ret))
		goto errt1;
	SM_TEST(evthr_ctx != NULL);

	ret = dns_rslv_new(strlen(host));
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret))
		goto errt;
	if (flags & T_OPT_DEBUG)
		_res.options |= RES_DEBUG;

	ret = dns_mgr_ctx_new(0, DNS_TIMEOUT, 0, 0, &dns_mgr_ctx);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret))
		goto err0;
	SM_TEST(dns_mgr_ctx != NULL);
	if (NULL == dns_mgr_ctx)
		goto err0;

	ret = dns_tsk_new(dns_mgr_ctx,
			(flags & T_OPT_USE_TCP) ? DNS_TSK_FL_USETCP : 0, ipv4, 0, &dns_tsk);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		sm_io_fprintf(smioout, "dns_tsk_new: ret=%x\n", ret);
		goto err1;
	}
	SM_TEST(dns_tsk != NULL);
	if (NULL == dns_tsk)
		goto err1;

	ret = dns_tsk_start(dns_mgr_ctx, evthr_ctx);
	SM_TEST(sm_is_success(ret));
	if (sm_is_err(ret))
		goto err1;

	type = T_A;
	if (flags & T_OPT_MX) {
		type = T_MX;
		ret = dns_req_add(dns_mgr_ctx, query, type, Timeout, callback, (void *) NULL);
		SM_TEST(sm_is_success(ret));
		if (sm_is_err(ret))
			goto err1;
	}

	if (flags & T_OPT_A) {
		type = T_A;
		ret = dns_req_add(dns_mgr_ctx, query, type, Timeout, callback, (void *) NULL);
		SM_TEST(sm_is_success(ret));
		if (sm_is_err(ret))
			goto err1;
	}

#if MTA_NETINET6
	if (flags & T_OPT_AAAA) {
		type = T_AAAA;
		ret = dns_req_add(dns_mgr_ctx, query, type, Timeout, callback, (void *) NULL);
		SM_TEST(sm_is_success(ret));
		if (sm_is_err(ret))
			goto err1;
	}
#endif

	/* test: add the same query again... */
	if (flags & T_OPT_TWICE) {
		ret = dns_req_add(dns_mgr_ctx, query, type, Timeout, callback, (void *) NULL);
		SM_TEST(sm_is_success(ret));
		if (sm_is_err(ret))
			goto err1;
	}

	n = gettimeofday(&sleept, NULL);
	SM_TEST(n == 0);
	sleept.tv_usec += 1000;
	ret = evthr_task_new(evthr_ctx, &task3, EVTHR_EV_SL, -1, &sleept,
				fcts, (void *) NULL);
	SM_TEST(sm_is_success(ret));
	SM_TEST(task3 != NULL);

	ret = evthr_loop(evthr_ctx);
	SM_TEST(sm_is_success(ret));

	/* need to get result... */

  err1:
	ret = dns_mgr_ctx_del(dns_mgr_ctx);
	SM_TEST(SM_SUCCESS == ret);
  err0:
  errt:
	ret = evthr_stop(evthr_ctx, 0);
	SM_TEST(sm_is_success(ret));
  errt1:
	ret = thr_stop();
	SM_TEST(sm_is_success(ret));
  errq:
	SM_CSTR_FREE(query);
	return;
}

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "Usage: %s [options] names\n", prg);
	sm_io_fprintf(smioerr, "Options:\n"
#if MTA_NETINET6
		"-A         query for AAAA record\n"
#endif
		"-a         query for A record\n"
		"-d         query twice\n"
		"-D         turn on debugging\n"
		"-i ip      use ip address for nameserver [127.0.0.1]\n"
		"-m         do not query for MX record\n"
		"-T         use TCP instead of UDP\n"
		"-t n       set MaxCount to n [%d] (wait 0.1s * MaxCount for results)\n"
		"-V         increase verbosity\n"
		"-w n       set DNS lookup timeout to n [%d]\n"
		, MaxCount, Timeout
		);
}

int
main(int argc, char *argv[])
{
	int r, i;
	uint flags;
	uint32_t ipv4;

	flags = T_OPT_MX;
	ipv4 = htonl(LOCALHOST_IP);
	while ((r = getopt(argc, argv, "AaCdDi:mt:TVw:")) != -1) {
		switch (r) {
		  case 'A':
#if MTA_NETINET6
			flags |= T_OPT_AAAA;
#endif
			break;
		  case 'a':
			flags |= T_OPT_A;
			break;
		  case 'C':
			compoptions(smioout);
			exit(0);
		  case 'd':
			flags |= T_OPT_TWICE;
			break;
		  case 'D':
			flags |= T_OPT_DEBUG;
			break;
		  case 'i':
			ipv4 = inet_addr(optarg);
			break;
		  case 'm':
			flags &= ~T_OPT_MX;
			break;
		  case 't':
			i = atoi(optarg);
			if (i <= 0)
				return EX_USAGE;
			MaxCount = (uint) i;
			break;
		  case 'T':
			flags |= T_OPT_USE_TCP;
			break;
		  case 'V':
			Verbose++;
			break;
		  case 'w':
			i = atoi(optarg);
			if (i < 0)
				return EX_USAGE;
			Timeout = i;
			break;
		  default:
			usage(argv[0]);
			return EX_USAGE;
		}
	}

	sm_test_begin(argc, argv, "test getmxrr");
	for (r = optind; r < argc; r++)
		testdns(argv[r], flags, ipv4);
	sm_io_flush(smioout);
	return sm_test_end();
}
