/*
 * Copyright (c) 2002-2006, 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: rsc.c,v 1.99 2014/01/30 16:37:29 ca Exp $")
#include "sm/limits.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/common.h"
#include "sm/memops.h"
#include "sm/io.h"
#include "sm/heap.h"
#include "sm/queue.h"
#include "sm/bhtable.h"

#define RSC_COMPILE 1

#if SM_RSC_TYPED
# include "sm/rsct.h"
#else
# include "sm/rsc.h"
#endif

#include "sm/pthread.h"

#ifndef SM_RSC_STATS
# define SM_RSC_STATS 1
#endif

/*
**  Notice:
**	- if locking is active (MTA_USE_PTHREADS) then each function locks
**	  the entire RSC during its operation. That means none of these
**	  functions can call another one herein (if both use locking)!
**	- copying keys currently requires that they are all the same size!!!
**	maybe define some size in the context? that would work well enough
**	for IPv{4,6}. then we also don't need to pass a flag to "add" but
**	use a flag in rsc itself(?)
**
**  Possible enhancements:
**	- change size of RSC? it should be easy to extend the RSC,
**	  there is a function for BHT to grow.  However, there is no
**	  function for BHT to shrink (which shouldn't matter much,
**	  we can just restrict the size of the RSC; that won't
**	  reclaim the memory in the BHT, but a bit in the RSC).
*/

/* abstraction layer for linking, XXX entries in struct are not abstracted */
#define link_P	rsc_entry_P
#define LINK_SUCC(r) TAILQ_NEXT(r, rsce_link)
#define LINK_PRED(r) TAILQ_PREV(r, rsc_hd, rsce_link)
#define LINK_INIT(rsc) TAILQ_INIT(&((rsc)->rsc_link))
#define LINK_APPEND(rsc, entry)	TAILQ_INSERT_HEAD(&((rsc)->rsc_link), entry, rsce_link)
#define LINK_REMOVE(rsc, entry) TAILQ_REMOVE(&((rsc)->rsc_link), (entry), rsce_link)
#define LINK_OLDEST(rsc)	TAILQ_LAST(&((rsc)->rsc_link), rsc_hd)
#define LINK_NEWEST(rsc)	TAILQ_FIRST(&((rsc)->rsc_link))
#define LINK_END(rsc)		TAILQ_END(&((rsc)->rsc_link))

/* minimum size of an RSC */
#define RSC_MIN_SIZE	5

typedef struct rsc_entry_S rsc_entry_T, *rsc_entry_P;

/*
**  RSC entries are kept in most-recently used order.
**  A BHT is used to find RSC entries.
*/

struct rsc_entry_S
{
	sm_magic_T	 sm_magic;
	TAILQ_ENTRY(rsc_entry_S)	 rsce_link;	/* MRU linkage */
#if SM_RSC_TYPED
	uint		 rsce_type;	/* type of entry */
#endif
	void		*rsce_value;	/* corresponding value */
	uint		 rsce_len;	/* key len */
	union
	{
		char	*rscu_key;		/* pointer to lookup key */
		char	 rscu_keyl[sizeof(void *)];	/* local copy of lookup key */
	} rsce_keyu;
#define rsce_key	rsce_keyu.rscu_key
#define RSCE_KEY(entry, flags)	(SM_IS_FLAG(flags, RSC_AFL_CPK) ? (entry)->rsce_keyu.rscu_keyl : (entry)->rsce_keyu.rscu_key)
};

#if SM_RSC_TYPED
# define RSC_DELETE(rsc, value, type) (rsc)->rsc_delete(value, (rsc)->rsc_ctx, (type))
# define RSCTYPE	, type
#else
# define RSC_DELETE(rsc, value, type) (rsc)->rsc_delete(value, (rsc)->rsc_ctx)
# define RSCTYPE
#endif

struct sm_rsc_S
{
	sm_magic_T	 sm_magic;
	bht_P		 rsc_table;				/* table with key, rsc_entry pairs */
	uint		 rsc_minsize;			/* size for copied key (if >0) */
	uint		 rsc_limit;	/* max # of entries */
	uint		 rsc_used;	/* current # of entries */
	TAILQ_HEAD(rsc_hd, rsc_entry_S)	 rsc_link;	/* MRU linkage */
	void		*rsc_ctx;				/* application context */
	sm_rsc_create_F	 rsc_create;		/* constructor */
	sm_rsc_delete_F	 rsc_delete;		/* destructor */
	rsc_entry_P	 rsc_cursorn;
#if MTA_USE_PTHREADS
	pthread_mutex_t	 rsc_mutex;
#endif
#if SM_RSC_STATS
	uint		 rsc_max_used;	/* max # of entries */
#endif
};

#define SM_IS_RSC(rsc)	SM_REQUIRE_ISA((rsc), SM_RSC_MAGIC)
#define SM_IS_RSCE(rsce)	SM_REQUIRE_ISA((rsce), SM_RSCE_MAGIC)

/*
**  SM_RSC_NEW -- create an RSC
**
**	Parameters:
**		limit -- maximum number of entries in the RSC
**		htsize -- size of BHT
**		create -- function to create an entry
**		delete -- function to delete an entry
**		context -- context for create/delete functions
**
**	Returns:
**		Pointer to RSC (NULL on error)
**		(fixme: should we return the usual error code??)
**
**	Last code review: 2005-04-07 22:20:17
**	Last code change: 2005-04-07 22:20:11
*/

sm_rsc_P
sm_rsc_new(uint limit, uint htsize, uint minsize, sm_rsc_create_F create, sm_rsc_delete_F delete, void *context)
{
	sm_rsc_P rsc;
	uint total_limit;
#if MTA_USE_PTHREADS
	int r;
#endif

	if (limit < 1)
		return NULL;
	rsc = (sm_rsc_P) sm_malloc(sizeof(*rsc));
	if (NULL == rsc)
		return NULL;
	total_limit = limit;
	if (htsize < total_limit)	/* this include htsize = 0 */
		htsize = total_limit;

#if MTA_USE_PTHREADS
	r = pthread_mutex_init(&rsc->rsc_mutex, SM_PTHREAD_MUTEXATTR);
	if (r != 0)
		return NULL; /* sm_error_perm(SM_EM_RSC, r); */
#endif
	rsc->rsc_table = bht_new(htsize, total_limit);
	if (NULL == rsc->rsc_table) {
		sm_free_size(rsc, sizeof(*rsc));
		goto error;
	}
	if (limit < RSC_MIN_SIZE)
		limit = RSC_MIN_SIZE;
	rsc->rsc_limit = limit;
	rsc->rsc_used = 0;
#if SM_RSC_STATS
	rsc->rsc_max_used = 0;
#endif

	rsc->rsc_minsize = minsize;
	rsc->rsc_create = create;
	rsc->rsc_delete = delete;
	LINK_INIT(rsc);
	rsc->rsc_ctx = context;
	rsc->rsc_cursorn = NULL;
	rsc->sm_magic = SM_RSC_MAGIC;
	return rsc;

  error:
#if MTA_USE_PTHREADS
	(void) pthread_mutex_destroy(&rsc->rsc_mutex);
#endif
	return NULL;
}

/*
**  SM_RSC_ADD -- add item to RSC or return old one
**
**	Parameters:
**		rsc -- RSC
**		flags -- flags
**		key -- key to lookup
**		len -- length of key
**		value -- value to add (passed to create function)
**		pvalue -- pointer to value of entry (output)
**		locktype -- kind of locking
**
**	Returns:
**		>=0: usage
**		<0: usual sm_error code; ENOMEM, SM_E_FULL
**
**	Side Effects: if RSC is full the function removes the oldest
**		entry which is not "undone" on error.
**
**	Locking: locks entire RSC during operation, returns unlocked
**		(depending on locktype)
**
**	Last code review: 2005-04-07 22:17:15; see comments
**	Last code change:
*/

sm_ret_T
sm_rsc_add(sm_rsc_P rsc, uint flags,
#if SM_RSC_TYPED
	uint type,
#endif
	char *key, uint len, void *value, void **pvalue
#if MTA_USE_PTHREADS
	, thr_lock_T locktype
#endif
	)
{
	rsc_entry_P entry;
	sm_ret_T ret;

	SM_IS_RSC(rsc);
	C_THREAD_LOCK(thr_lock_it(locktype), &rsc->rsc_mutex);
	entry = (rsc_entry_P) bht_find(rsc->rsc_table, key, len);
	if (NULL != entry)
		SM_IS_RSCE(entry);
	if (NULL == entry) {
		bool allocated;
		bht_entry_P bht_entry;

		allocated = false;

		/* not in RSC: is limit reached? */
		if (rsc->rsc_used >= rsc->rsc_limit) {

			if (!SM_IS_FLAG(flags, RSC_AFL_DELOK)) {
				ret = sm_error_temp(SM_EM_RSC, SM_E_FULL);
				goto error;
			}

			/* remove oldest entry (not "undone" on error!) */
			entry = LINK_OLDEST(rsc);
			SM_IS_RSCE(entry);
			if (rsc->rsc_delete != NULL) {
				ret = RSC_DELETE(rsc, entry->rsce_value, entry->rsce_type);
				if (sm_is_err(ret))
					goto error;
			}
			LINK_REMOVE(rsc, entry);
			bht_rm(rsc->rsc_table, RSCE_KEY(entry, flags), entry->rsce_len,
					NULL, NULL);
			if (SM_IS_FLAG(flags, RSC_AFL_CPK) && len > sizeof(void *)
				&& entry->rsce_len < len)
			{
				/* XXX need to resize! */
				SM_ASSERT(!(SM_IS_FLAG(flags, RSC_AFL_CPK)
					&& len > sizeof(void *)
					&& entry->rsce_len < len));
			}
		}
		else {
			size_t el;

			el = sizeof(*entry);
			if (SM_IS_FLAG(flags, RSC_AFL_CPK)) {
				if (rsc->rsc_minsize > 0) {
					SM_REQUIRE(len <= rsc->rsc_minsize);
					el += rsc->rsc_minsize - sizeof(void *);
				}
				else if (len > sizeof(void *))
					el += len - sizeof(void *);
			}
			entry = (rsc_entry_P) sm_zalloc(el);
			if (NULL == entry) {
				ret = sm_error_temp(SM_EM_RSC, ENOMEM);
				goto error;
			}
			entry->sm_magic = SM_RSCE_MAGIC;
			allocated = true;
			rsc->rsc_used++;
#if SM_RSC_STATS
			if (rsc->rsc_used > rsc->rsc_max_used)
				rsc->rsc_max_used = rsc->rsc_used;
#endif

		}

		/* this must not fail... fixme: handle NULL return? */
		entry->rsce_value = rsc->rsc_create(key, len, value,
						rsc->rsc_ctx RSCTYPE);
		if (SM_IS_FLAG(flags, RSC_AFL_CPK))
			sm_memcpy(entry->rsce_keyu.rscu_keyl, key, len);
		else
			entry->rsce_key = key;
		entry->rsce_len = len;
		ret = bht_add(rsc->rsc_table, RSCE_KEY(entry, flags), len,
				(char *) entry, &bht_entry);
		if (sm_is_err(ret)) {
			/* clean up! */
			if (rsc->rsc_delete != NULL)
				RSC_DELETE(rsc, entry->rsce_value, entry->rsce_type);
			if (allocated) {
				entry->sm_magic = SM_MAGIC_NULL;
				sm_free(entry);
			}
			rsc->rsc_used--;
			goto error;
		}
#if SM_RSC_TYPED
		entry->rsce_type = type;
#endif

		/* insert entry right at (after) the head */
		LINK_APPEND(rsc, entry);
	}
	else {
		if (SM_IS_FLAG(flags, RSC_AFL_UPD) && entry->rsce_value != value) {
			RSC_DELETE(rsc, entry->rsce_value, entry->rsce_type);
			entry->rsce_value = rsc->rsc_create(key, len, value,
						rsc->rsc_ctx RSCTYPE);
		}

		if (entry != LINK_NEWEST(rsc)) {
			/* move the entry up front */
			LINK_REMOVE(rsc, entry);
			LINK_APPEND(rsc, entry);
		}
	}
	if (pvalue != NULL)
		*pvalue = entry->rsce_value;
	C_THREAD_UNLOCK(thr_unl_no_err(locktype), &rsc->rsc_mutex);
	return (int) ((rsc->rsc_used * 100) / rsc->rsc_limit);

  error:
	C_THREAD_UNLOCK(thr_unl_if_err(locktype), &rsc->rsc_mutex);
	return ret;
}

/*
**  SM_RSC_RM -- remove item from RSC.
**
**	Parameters:
**		rsc -- RSC
**		key -- key to lookup
**		len -- length of key
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code; SM_E_NOTFOUND
**
**	Side Effects: no side effects on error.
**
**	Locking: locks entire RSC during operation, returns unlocked
**		(depending on locktype)
**
**	Last code review: 2005-03-31 17:51:39
**	Last code change:
*/

sm_ret_T
sm_rsc_rm(sm_rsc_P rsc, const char *key, uint len
#if MTA_USE_PTHREADS
	, thr_lock_T locktype
#endif
	)
{
	sm_ret_T ret;
	rsc_entry_P entry;

	SM_IS_RSC(rsc);
	C_THREAD_LOCK(thr_lock_it(locktype), &rsc->rsc_mutex);
	ret = SM_SUCCESS;
	entry = (rsc_entry_P) bht_find(rsc->rsc_table, key, len);
	if (NULL == entry) {
		ret = sm_error_perm(SM_EM_RSC, SM_E_NOTFOUND);
		goto error;
	}
	SM_IS_RSCE(entry);
	if (entry == rsc->rsc_cursorn)
		rsc->rsc_cursorn = LINK_SUCC(rsc->rsc_cursorn);

	LINK_REMOVE(rsc, entry);
	if (rsc->rsc_delete != NULL)
		RSC_DELETE(rsc, entry->rsce_value, entry->rsce_type);
	bht_rm(rsc->rsc_table, key, len, NULL, NULL);
	SM_ASSERT(rsc->rsc_used > 0);
	--rsc->rsc_used;
	entry->sm_magic = SM_MAGIC_NULL;
	sm_free(entry);
	C_THREAD_UNLOCK(thr_unl_no_err(locktype), &rsc->rsc_mutex);
	return ret;

  error:
	C_THREAD_UNLOCK(thr_unl_if_err(locktype), &rsc->rsc_mutex);
	return ret;
}

/*
**  SM_RSC_LOOKUP -- look up item in RSC
**
**	Parameters:
**		rsc -- RSC
**		key -- key to lookup
**		len -- length of key
**		mru -- "touch" item such that it is most recently used?
**		locktype -- kind of locking
**
**	Returns:
**		Pointer to value of entry (NULL if not found)
**		(fixme: should we return the usual error code??)
**
**	Locking: locks entire RSC during operation, returns unlocked
**		(depending on locktype)
**		Note: the item that is returned is not locked hence
**		there's a possible race condition depending on the
**		application (how it uses those items).
**
** #if SM_RSC_TYPED
**	Restrict this to a certain type?
** #endif
**
**	Last code review: 2005-03-31 17:47:15
**	Last code change:
*/

const void *
sm_rsc_lookup(sm_rsc_P rsc, const char *key, uint len, bool mru
#if MTA_USE_PTHREADS
	, thr_lock_T locktype
#endif
	)
{
	rsc_entry_P entry;

	SM_IS_RSC(rsc);
	C_THREAD_LOCK(thr_lock_it(locktype), &rsc->rsc_mutex);
	entry = (rsc_entry_P) bht_find(rsc->rsc_table, key, len);
	if (mru && NULL != entry && entry != LINK_NEWEST(rsc)) {
		/* move the entry up front */
		LINK_REMOVE(rsc, entry);
		LINK_APPEND(rsc, entry);
	}
	C_THREAD_UNLOCK(thr_unl_always(locktype), &rsc->rsc_mutex);
	if (NULL == entry)
		return NULL;
	SM_IS_RSCE(entry);
	return entry->rsce_value;
}

/*
**  SM_RSC_FREE_CALLBACK -- callback function for rsc_free
**
**	Parameters:
**		ptr -- pointer to RSC element
**		key -- key (unused)
**		ctx -- context.
**
**	Returns:
**		Nothing.
**
**	Last code review: 2005-04-07 22:19:13
**	Last code change: 2005-04-07 22:18:49
*/

static void
sm_rsc_free_callback(void *ptr, void *key, void *ctx)
{
	rsc_entry_P entry;
	sm_rsc_P rsc;

	(void) key;
	SM_REQUIRE(ptr != NULL);
	entry = (rsc_entry_P) ptr;
	SM_IS_RSCE(entry);
	rsc = (sm_rsc_P) ctx;
	SM_IS_RSC(rsc);
	if (rsc->rsc_delete != NULL)
		RSC_DELETE(rsc, entry->rsce_value, entry->rsce_type);
	entry->sm_magic = SM_MAGIC_NULL;
	sm_free(entry);
}

/*
**  SM_RSC_FREE -- free rsc and its contents
**
**	Parameters:
**		rsc -- RSC
**
**	Returns:
**		Nothing.
**
**	Last code review: 2005-04-07 22:22:52
**	Last code change: 2005-04-07 22:22:48
*/

void
sm_rsc_free(sm_rsc_P rsc)
{
#if MTA_USE_PTHREADS
	int r;
#endif

	SM_IS_RSC(rsc);
	bht_destroy(rsc->rsc_table, sm_rsc_free_callback, (void *) rsc);
#if MTA_USE_PTHREADS
	r = pthread_mutex_destroy(&rsc->rsc_mutex);
	if (r != 0) {
		/* LOG an error?? */
	}
#endif
	rsc->sm_magic = SM_MAGIC_NULL;
	sm_free_size(rsc, sizeof(*rsc));
}

/*
**  SM_RSC_WALK -- iterate over all rsc entries
**
**	Parameters:
**		rsc -- RSC
**		action -- function to apply
**		ctx -- context for action
**		locktype -- kind of locking
**
**	Returns:
**		SM_SUCCESS except for (un)lock errors
**
**	Locking: locks entire RSC during operation, returns unlocked
**		(depending on locktype)
**
**	Last code review: 2005-04-07 22:34:25
**	Last code change:
*/

sm_ret_T
sm_rsc_walk(sm_rsc_P rsc, sm_rsc_walk_F *action, void *ctx
#if MTA_USE_PTHREADS
	, thr_lock_T locktype
#endif
	)
{
	link_P entry;

	SM_IS_RSC(rsc);
	SM_REQUIRE(action != NULL);
	C_THREAD_LOCK(thr_lock_it(locktype), &rsc->rsc_mutex);
	for (entry = LINK_NEWEST(rsc); entry != LINK_END(rsc); entry = rsc->rsc_cursorn)
	{
		SM_IS_RSCE(entry);
		rsc->rsc_cursorn = LINK_SUCC(entry);
		action(entry->rsce_key, entry->rsce_value, rsc->rsc_ctx, ctx
#if SM_RSC_TYPED
			, entry->rsce_type
#endif
			);
	}
	C_THREAD_UNLOCK(thr_unl_always(locktype), &rsc->rsc_mutex);
	rsc->rsc_cursorn = NULL;
	return SM_SUCCESS;
}

/*
**  SM_RSC_WALK_R -- iterate over rsc entries
**
**	Parameters:
**		rsc -- RSC
**		action -- function to apply
**		ctx -- context for action
**		locktype -- kind of locking
**
**	Returns:
**		SM_SUCCESS except for (un)lock errors
**
**	Locking: locks entire RSC during operation, returns unlocked
**		(depending on locktype)
**
**	Last code review: 2005-04-07 22:34:25
**	Last code change:
*/

sm_ret_T
sm_rsc_walk_r(sm_rsc_P rsc, sm_rsc_walk_r_F *action, void *ctx
#if MTA_USE_PTHREADS
	, thr_lock_T locktype
#endif
	)
{
	sm_ret_T ret;
	link_P entry;

	SM_IS_RSC(rsc);
	SM_REQUIRE(action != NULL);
	ret = SM_SUCCESS;
	C_THREAD_LOCK(thr_lock_it(locktype), &rsc->rsc_mutex);
	for (entry = LINK_OLDEST(rsc); entry != LINK_END(rsc); entry = rsc->rsc_cursorn)
	{
		SM_IS_RSCE(entry);
		rsc->rsc_cursorn = LINK_PRED(entry);
		ret = action(entry->rsce_key, entry->rsce_value, rsc->rsc_ctx, ctx
#if SM_RSC_TYPED
			, entry->rsce_type
#endif
			);
		if (ret != SM_SUCCESS)
			break;
	}
	C_THREAD_UNLOCK(thr_unl_always(locktype), &rsc->rsc_mutex);
	rsc->rsc_cursorn = NULL;
	return ret;
}


/*
**  SM_RSC_USAGE -- return percentage of # of item in RSC (* 100)
**
**	Parameters:
**		rsc -- RSC
**
**	Returns:
**		fill level
**
**	Locking: no locking, may return garbage.
**
**	Last code review: 2005-04-07 22:34:33
**	Last code change:
*/

uint
sm_rsc_usage(sm_rsc_P rsc)
{
	SM_IS_RSC(rsc);
	return (uint) ((rsc->rsc_used * 100) / rsc->rsc_limit);
}

/*
**  SM_RSC_ENTRIES -- return # of items in RSC
**
**	Parameters:
**		rsc -- RSC
**		pmax_used -- (pointer to) max. number of entries (output)
**
**	Returns:
**		fill level
**
**	Locking: no locking, may return garbage.
**
**	Last code review: 2005-04-07 22:34:40
**	Last code change: 2005-05-19 20:16:33
*/

uint
sm_rsc_entries(sm_rsc_P rsc, uint *pmax_used)
{
	SM_IS_RSC(rsc);
	if (pmax_used != NULL)
#if SM_RSC_STATS
		*pmax_used = rsc->rsc_max_used;
#else
		*pmax_used = 0;
#endif
	return rsc->rsc_used;
}

/*
**  SM_RSC_AVAIL -- return # of available items in RSC
**
**	Parameters:
**		rsc -- RSC
**
**	Returns:
**		number of available entries
**
**	Locking: no locking, may return garbage.
*/

uint
sm_rsc_avail(sm_rsc_P rsc)
{
	SM_IS_RSC(rsc);
	return rsc->rsc_limit - rsc->rsc_used;
}

/*
**  SM_RSC_STATS -- return bht stats of RSC
**
**	Parameters:
**		rsc -- RSC
**		out -- buffer to store stats in
**		len -- length of buffer out
**
**	Returns:
**		none
**
**	Locking: no locking, may return garbage.
**
**	Last code review: 2005-04-07 22:34:46
**	Last code change:
*/

void
sm_rsc_stats(sm_rsc_P rsc, char *out, uint len)
{
	uint l;
	char buf[256];

	SM_IS_RSC(rsc);
	l = 0;
	(void) sm_snprintf(buf, sizeof(buf),
		"limit=%4u\n"
		"used =%4u\n"
#if SM_RSC_STATS
		"max  =%4u\n"
#endif
		, rsc->rsc_limit
		, rsc->rsc_used
#if SM_RSC_STATS
		, rsc->rsc_max_used
#endif
		);
	l = strlcat(out, buf, len);
	if (l >= len)
		return;
	if (len > l)
		bht_stats(rsc->rsc_table, out + l, len - l);
}
